#include "msp430.hpp"

static int ana_basic(void);

size_t cur_cmd_size;

static ushort get_next_word( void )
{
	cur_cmd_size += 2;
	uchar lsb = ua_next_byte();
	uchar msb = ua_next_byte();
	return (lsb) | (msb << 8);
}

// Immediate Mode #N
// The word following the instruction contains the immediate constant N
static void opImmed ( op_t &op )
{
	op.type  = o_imm;
	op.offb  = cur_cmd_size;
	op.value = get_next_word();
	TRACE("Imm[%X]", op.value);
}

// Immediate Mode #N from Constant Generator
// immediate constant N get from CG2 
static void opImmedCG2 ( op_t &op, ushort imm )
{
	op.type    = o_imm;
	op.value   = imm;
	op.specval |= OPSPECVAL_CG_FLAG;
	TRACE("ImmConst[%X]", op.value);
}

// Register Mode Rn
// Register contents are operand
static void opReg( op_t &op, uchar reg )
{
	op.type = o_reg;
	op.reg  = reg;
	TRACE("%s", ph.regNames[reg]);
}

// Indirect Register Mode @Rn
// Rn is used as a pointer to the operand
static void opRegIndirect( op_t &op, uchar reg )
{
	op.type   = o_phrase;
	op.phrase = phrase_ind_reg;
	op.indreg = reg;
	TRACE("@%s", ph.regNames[reg]);
}

// Indirect Autoincrement @Rn+
// Rn is used as a pointer to the operand. Rn is incremented afterwards
static void opRegIndAutoInc ( op_t &op, uchar reg )
{
	op.type   = o_phrase;
	op.phrase = phrase_ind_reg_inc;
	op.indreg = reg;
	TRACE("@%s++", ph.regNames[reg]);
}

// Indexed Mode X(Rn)
// (Rn + X) points to the operand. X is stored in the next word
static void opIndexed( op_t &op, uchar reg )
{
	op.offb  = cur_cmd_size;
	ushort x = get_next_word();
	op.type = o_displ;
	op.reg  = reg;
	op.addr = x;
	TRACE("%X(%s)", x, ph.regNames[reg]);
}

// Absolute Mode &ADDR
// The word following the instruction contains the absolute address.
static void opAbsolute( op_t &op )
{
	op.offb  = cur_cmd_size;
	ushort addr = get_next_word();
	op.type = o_mem;
	op.addr = addr;
	TRACE("&%X", addr);
}

// Symbolic Mode ADDR
//(PC + X) points to the operand. X is stored in the next word. Indexed Mode X(PC) is used
static void opSymbolic( op_t &op )
{
	op.type = o_near;
	op.offb = cur_cmd_size;
	op.addr = (get_next_word() + cmd.ip + cur_cmd_size - 2) & 0xFFFF;
	TRACE("%X", op.addr);
}

//----------------------------------------------------------------------
// Parse source operand
static void parseSrcOperand ( op_t &op, uchar as, uchar reg )
{
	if (reg == rCG2) { 
		// source is constant generator rCG2 (0, 1, 2, -1)
		const ushort imms[] = {0, 1, 2, 0xFFFF};
		opImmedCG2(op, imms[as >> 4]);
		return;
	}

	if ((reg == rSR) && ((as == (2<<4)) || (as == (3<<4)))) {
		opImmedCG2(op, (as == (2<<4)) ? 4 : 8 );
		return;
	}

	switch (as)
	{
	case (0 << 4):
		opReg(op, reg); // operand is register
		break;
	case (1 << 4):
		switch(reg)
		{
		case rSR: opAbsolute(op);     break;
		case 0:   opSymbolic(op);     break;
		default:  opIndexed(op, reg); break;
		}
		break;
	case (2 << 4):
		opRegIndirect(op, reg); // operand is register indirect
		break;
	case (3 << 4):
		if (reg == 0) {
			opImmed(op);        // operand is immediate
		} else {
			opRegIndAutoInc(op, reg);
		}
		break;
	}
}

//----------------------------------------------------------------------
static void parseDstOperand( op_t &op, bool as, uchar reg )
{
	if (reg == rCG2) {
		// constant generator rCG2 (0, 1, 2, -1)
		const ushort imms[] = {0, 1, 2, 0xFFFF};
		opImmedCG2(op, as ? 1 : 0 );
		return;
	}

	if (as) {
		switch(reg)
		{
		case rSR: opAbsolute(op);     break;
		case 0:   opSymbolic(op);     break;
		default:  opIndexed(op, reg); break;
		}
	} else {
		opReg(op, reg); // operand is register
	}
}

//----------------------------------------------------------------------
// Parse instruction format 1
// Op1 - source
// Op2 - destination
static void instruction_fmt_1( ushort code )
{
	cur_cmd_size = 2;

	TRACE("%s ", Instructions[cmd.itype].name);

	// parse source
	TRACE("src:");
	uchar as  = (code & ((1<<4)|(1<<5))) & 0xFF;
	uchar reg = (code & 0x0F00) >> 8;
	parseSrcOperand(cmd.Op1, as, reg);

	// parse destination
	TRACE(", dst:");
	as  = (code & (1<<7)) & 0xFF;
	reg = code & (0x0F);
	parseDstOperand(cmd.Op2, (as != 0), reg);

	cmd.size = cur_cmd_size;
	TRACE(" size:%d bytes", cmd.size);
}

//----------------------------------------------------------------------
// Parse instruction format 2
// Op1 - source
// Op2 - destination
static void instruction_fmt_2( ushort code )
{
	cur_cmd_size = 2;
	TRACE("%s ", Instructions[cmd.itype].name);
	uchar as  = (code & ((1<<4)|(1<<5))) & 0xFF;
	uchar reg = code & (0x0F);
	parseSrcOperand(cmd.Op2, as, reg);
	cmd.size = cur_cmd_size;
	TRACE(" size:%d bytes", cmd.size);
}

//----------------------------------------------------------------------
// Parse instruction format 3
// Op1 - source
// Op2 - destination
static void instruction_fmt_3( ushort code )
{
	TRACE("%s ", Instructions[cmd.itype].name);
	cmd.Op2.type = o_near;
	ulong addr = cmd.ip + 2;
	ulong ofs  = (code & 0x1FF); 
	if (code & (1<<9)) ofs |= ~(0x1FF);
	cmd.Op2.addr = (long)addr + (long)(ofs*2);
	TRACE("%X", cmd.Op2.addr);
	cmd.size  = 2;
	TRACE(" size:%d bytes", cmd.size);
}

//----------------------------------------------------------------------
// Mark instruction byte access flag
inline void mark_byte_access ( void )
{
	cmd.auxpref = CMDAUXPREF_BYTE_FLAG;
}

//----------------------------------------------------------------------
// Check instruction for convert to RL
static bool is_operand_equal_to_RLx( void )
{
	if (cmd.Op1.type != cmd.Op2.type) return false;
	if ((cmd.Op1.type == o_phrase) || (cmd.Op1.type == o_imm)) return false;
	if (cmd.Op1.reg  != cmd.Op2.reg)  return false;
	return (cmd.Op1.addr == cmd.Op2.addr);
}

//----------------------------------------------------------------------
// Check operand for immediate from Constant Generator
inline bool is_operand_imm_cg ( op_t &op )
{
	return ((op.type == o_imm) && (op.specval & OPSPECVAL_CG_FLAG));
}

//----------------------------------------------------------------------
// Check operand for register
inline bool is_operand_reg ( op_t &op, ushort reg )
{
	return ((o_reg == op.type) && (reg == op.reg));
}

//----------------------------------------------------------------------
// Copy src operand to dst.
static void copy_operand ( op_t &dst, op_t &src )
{
	dst.type      = src.type;
	dst.offb	  = src.offb;
	dst.offo	  = src.offo;
	dst.flags	  = src.flags;
	dst.dtyp	  = src.dtyp;
	dst.reg	      = src.reg;
	dst.value	  = src.value;
	dst.addr	  = src.addr;
	dst.specval	  = src.specval;
	dst.specflag1 = src.specflag1;
	dst.specflag2 = src.specflag2;
	dst.specflag3 = src.specflag3;
	dst.specflag4 = src.specflag4;
}


//----------------------------------------------------------------------
// Convert basic instruction to emulate
void ana_create_emulate_instructions ( void )
{
	switch(cmd.itype)
	{
	case MSP430_MOV:
		if ( (cmd.Op1.type == o_phrase) && (cmd.Op1.phrase == phrase_ind_reg_inc) && (cmd.Op1.indreg == rSP) ) {
			// mov @SP+, 
			if (is_operand_reg(cmd.Op2, rPC)) {
				// MOV @SP+,PC -> RET
				cmd.itype = MSP430_EMU_RET;
				cmd.Op2.type = o_void;
			} else {
				// MOV @SP+,dst -> POP dst
				cmd.itype = MSP430_EMU_POP;
			}
			cmd.Op1.type = o_void;
			break;
		}
		if (is_operand_reg(cmd.Op2, rPC)) {
			// MOV dst,PC -> BR dst
			copy_operand(cmd.Op2, cmd.Op1);
			cmd.itype = MSP430_EMU_BR;
			cmd.Op1.type = o_void;
			break;
		}

		if ( (is_operand_imm_cg(cmd.Op1)) && (0 == cmd.Op1.value) ) {
			// src = 0
			if (is_operand_imm_cg(cmd.Op2)) {
				// MOV #0,CG -> NOP
				cmd.Op2.type = o_void;
				cmd.itype = MSP430_EMU_NOP;
			} else {
				// MOV #0,dst -> CLR dst
				cmd.itype = MSP430_EMU_CLR;
			}
			cmd.Op1.type = o_void;
		}
		break;

	case MSP430_RETI:
		cmd.Op2.type = o_void;
		break;

	case MSP430_ADD:
		if (is_operand_equal_to_RLx()) {
			// ADD dst,dst -> RLA dst
			cmd.itype = MSP430_EMU_RLA;
			cmd.Op1.type = o_void;
			break;
		}
		if ( is_operand_imm_cg(cmd.Op1) ) {
			if (1 == cmd.Op1.value) {
				// ADD dst,#1 -> INC dst
				cmd.itype = MSP430_EMU_INC;
				cmd.Op1.type = o_void;
				break;
			}
			if (2 == cmd.Op1.value) {
				// ADD dst,#2 -> INCD dst
				cmd.itype = MSP430_EMU_INCD;
				cmd.Op1.type = o_void;
				break;
			}
		}
		break;

	case MSP430_ADDC:
		if (is_operand_equal_to_RLx()) {
			// ADDC dst,dst -> RLC dst
			cmd.itype = MSP430_EMU_RLC;
			cmd.Op1.type = o_void;
			break;
		}
		if (is_operand_imm_cg(cmd.Op1)) {
			if (0 == cmd.Op1.value) {
				// ADDC dst,#0 -> ADC dst
				cmd.itype = MSP430_EMU_ADC;
				cmd.Op1.type = o_void;
				break;
			}
		}
		break;

	case MSP430_SUB:
		if (is_operand_imm_cg(cmd.Op1)) {
			if (1 == cmd.Op1.value) {
				// SUB dst,#1 -> DEC dst
				cmd.itype = MSP430_EMU_DEC;
				cmd.Op1.type = o_void;
				break;
			}
			if (2 == cmd.Op1.value) {
				// SUB dst,#2 -> DECD dst
				cmd.itype = MSP430_EMU_DECD;
				cmd.Op1.type = o_void;
				break;
			}
		}
		break;

	case MSP430_SUBC:
		if ( is_operand_imm_cg(cmd.Op1) && (0 == cmd.Op1.value) ) {
			// SUBC dst,#0 -> SBC dst
			cmd.itype = MSP430_EMU_SBC;
			cmd.Op1.type = o_void;
		}
		break;

	case MSP430_DADD:
		if ( is_operand_imm_cg(cmd.Op1) && (0 == cmd.Op1.value) ) {
			// DADC dst,#0 -> DADC dst
			cmd.itype = MSP430_EMU_DADC;
			cmd.Op1.type = o_void;
		}
		break;

	case MSP430_XOR:
		if ( is_operand_imm_cg(cmd.Op1) && (0xFFFF == cmd.Op1.value) ) {
			// XOR dst,#-1 -> INV dst
			cmd.itype = MSP430_EMU_INV;
			cmd.Op1.type = o_void;
		}
		break;

	case MSP430_CMP:
		if ( is_operand_imm_cg(cmd.Op1) && (0 == cmd.Op1.value) ) {
			// CMP dst,#0 -> TST dst
			cmd.itype = MSP430_EMU_TST;
			cmd.Op1.type = o_void;
		}
		break;

	case MSP430_BIS:
	case MSP430_BIC:
		if ( is_operand_imm_cg(cmd.Op1) && is_operand_reg(cmd.Op2, rSR) ) {
			if (1 == cmd.Op1.value) {
				// BIS #1,SR -> SETC
				cmd.itype = (MSP430_BIC == cmd.itype) ? MSP430_EMU_CLRC : MSP430_EMU_SETC;
				cmd.Op1.type = o_void;
				cmd.Op2.type = o_void;
				break;
			}
			if (2 == cmd.Op1.value) {
				// BIS #2,SR -> SETZ
				cmd.itype = (MSP430_BIC == cmd.itype) ? MSP430_EMU_CLRZ : MSP430_EMU_SETZ;
				cmd.Op1.type = o_void;
				cmd.Op2.type = o_void;
				break;
			}
			if (4 == cmd.Op1.value) {
				// BIS #4,SR -> SETN
				cmd.itype = (MSP430_BIC == cmd.itype) ? MSP430_EMU_CLRN : MSP430_EMU_SETN;
				cmd.Op1.type = o_void;
				cmd.Op2.type = o_void;
				break;
			}
			if (8 == cmd.Op1.value) {
				// BIS #8,SR -> EINT
				cmd.itype = (MSP430_BIC == cmd.itype) ? MSP430_EMU_DINT : MSP430_EMU_EINT;
				cmd.Op1.type = o_void;
				cmd.Op2.type = o_void;
				break;
			}
		}
		break;
	}
}

//----------------------------------------------------------------------
// Analyze a basic instruction
static int ana_basic( void )
{
	ushort code0 = ua_next_byte();
	code0 |= (ua_next_byte() << 8);

	TRACE("pc:%.4X: %.4X..  ", cmd.ip, code0);

	uchar high_nible = (code0 >> 8) & 0xF0;
	bool success;

	if (0x10 == high_nible) {
		success = true;
		switch (code0 & 0xFF80)
		{
		case 0x1040:
			mark_byte_access();
		case 0x1000:
			cmd.itype = MSP430_RRC;
			break;

		case 0x1140:
			mark_byte_access();
		case 0x1100:
			cmd.itype = MSP430_RRA;
			break;

		case 0x1240: 
			mark_byte_access();
		case 0x1200:
			cmd.itype = MSP430_PUSH;
			break;

		case 0x1080: cmd.itype = MSP430_SWPB; break;
		case 0x1180: cmd.itype = MSP430_SXT;  break;
		case 0x1280: cmd.itype = MSP430_CALL; break;
		case 0x1300: cmd.itype = MSP430_RETI; break;
		default:     success = false;         break;
		}
		if (success) {
			instruction_fmt_2(code0);
			TRACE(";\n");
		} else
			TRACE("fmt2 failed!\n");
	} 

	if ((high_nible == 0x20) || (high_nible == 0x30) ) {
		success = true;
		switch((code0 >> 8) & 0xFC)
		{
		case 0x20: cmd.itype = MSP430_JNE; break;
		case 0x24: cmd.itype = MSP430_JEQ; break;
		case 0x28: cmd.itype = MSP430_JNC; break;
		case 0x2C: cmd.itype = MSP430_JC;  break;
		case 0x30: cmd.itype = MSP430_JN;  break;
		case 0x34: cmd.itype = MSP430_JGE; break;
		case 0x38: cmd.itype = MSP430_JL;  break;
		case 0x3C: cmd.itype = MSP430_JMP; break;
		default:   success = false;        break;
		}
		if (success) {
			instruction_fmt_3(code0);
			TRACE(";\n");
		} else
			TRACE("fmt3 failed\n");
	}

	if ( (high_nible >= 0x40) && (high_nible <= 0xF0) ) {
		success = true;		
		switch((code0 >> 8) & 0xF0)
		{
		case 0x40: cmd.itype = MSP430_MOV;  break;
		case 0x50: cmd.itype = MSP430_ADD;  break;
		case 0x60: cmd.itype = MSP430_ADDC; break;
		case 0x70: cmd.itype = MSP430_SUBC; break;
		case 0x80: cmd.itype = MSP430_SUB;  break;
		case 0x90: cmd.itype = MSP430_CMP;  break;
		case 0xA0: cmd.itype = MSP430_DADD; break;
		case 0xB0: cmd.itype = MSP430_BIT;  break;
		case 0xC0: cmd.itype = MSP430_BIC;  break;
		case 0xD0: cmd.itype = MSP430_BIS;  break;
		case 0xE0: cmd.itype = MSP430_XOR;  break;
		case 0xF0: cmd.itype = MSP430_AND;  break;
		default:   success = false;         break;
		}
		if (success) {
			if (code0 & (1<<6)) mark_byte_access();
			instruction_fmt_1(code0);
			TRACE(";\n");
		} else
			TRACE("fmt1 failed\n");
	}	

	ana_create_emulate_instructions();

	if ( (cmd.itype == MSP430_CALL) || (cmd.itype == MSP430_EMU_BR) ) {
		if ((cmd.Op2.type == o_imm)) {
			cmd.Op2.addr = cmd.Op2.value;
			cmd.Op2.type = o_near;
		}
	}

	return cmd.size;
}

//----------------------------------------------------------------------
// Is instruction is code reference
bool is_branch_instruction (ushort type)
{
	return ( (MSP430_CALL == type) || (MSP430_EMU_BR == type) || 
		( (type >= MSP430_JEQ) && (type <= MSP430_JMP) ) );
}

//----------------------------------------------------------------------
void operand_make_void ( op_t &op )
{
	op.dtyp    = dt_word;
	op.addr    = 0;
	op.value   = 0;
	op.offb    = 0;
	op.offo    = 0;
	op.reg     = 0;
	op.specval = 0;
}

//----------------------------------------------------------------------
// Analyze an instruction
int ana( void )
{
	cmd.itype   = o_void;
	cmd.auxpref = 0;
	operand_make_void(cmd.Op1);
	operand_make_void(cmd.Op2);
	operand_make_void(cmd.Op3);

	return ana_basic();
}

//--------------------------------------------------------------------------
void interr( const char *module )
{
	const char *name = NULL;
	if ( cmd.itype < ph.instruc_end )
		name = Instructions[cmd.itype].name;
	else
		cmd.itype = ph.instruc_start;
	warning("%a(%s): internal error in %s", cmd.ea, name, module);
}