#include "msp430.hpp"

static bool flow;

//------------------------------------------------------------------------
static void doImmdValue( const op_t &op )
{
	doImmd(cmd.ea);
	switch (cmd.itype)
	{
	case MSP430_AND:
	case MSP430_XOR:
	case MSP430_BIC:
	case MSP430_BIS:
	case MSP430_BIT:
		op_num(cmd.ea, 1);
		break;
	}
	uFlag = getFlags(cmd.ea);             // refresh uFlag with new flags
}

//------------------------------------------------------------------------
ea_t calc_mem(ea_t ea)
{
	return toEA(cmd.cs, ea);
}

//------------------------------------------------------------------------
static void handle_operand( op_t &op, int loading )
{
	ea_t ea;

	switch(op.type)
	{

	case o_phrase:						// no special handling for these types
	case o_reg:
		break;

	case o_imm:                         // an immediate number as an operand
		if (!loading) interr("emu");
		doImmdValue(op);                // handle immediate number
		// if the value was converted to an offset, then create a data xref:
		if (isOff(uFlag, op.n)) ua_add_off_drefs(op, dr_O);
		break;

	case o_displ:
		doImmdValue(op);                // handle immediate number
		// if the value was converted to an offset, then create a data xref:
		if (isOff(uFlag, op.n)) ua_add_off_drefs(op, (loading) ? dr_R : dr_W);
		break;

	case o_near:                        // a code reference
		if (is_branch_instruction(cmd.itype)) {
			ea = calc_mem(op.addr);
			int iscall = InstrIsSet(cmd.itype, CF_CALL);
			ua_add_cref(op.offb, ea, iscall ? fl_CN : fl_JN);
			if ( flow && iscall ) {
				func_t *pfn = get_func(ea);
				if ( pfn != NULL && (pfn->flags & FUNC_NORET) ) flow = false;
			}
			break;
		}
		//  !! if not branch process as mem reference
		//	\/
	case o_mem:                         // an ordinary memory data reference
		ea = calc_mem(op.addr);
		ua_dodata(ea, (cmd.auxpref & CMDAUXPREF_BYTE_FLAG) ? dt_byte : dt_word);
		if (!loading) doVar(ea);        // write access
		ua_add_dref(op.offb, ea, loading ? dr_R : dr_W);
		break;

	default:
		interr("emu");
		break;
	}
}

//----------------------------------------------------------------------
// Emulate an instruction
// This function should:
//      - create all xrefs from the instruction
//      - perform any additional analysis of the instruction/program
//        and convert the instruction operands, create comments, etc.
//      - create stack variables
//      - analyse the delayed branches and similar constructs
// The kernel calls ana() before calling emu(), so you may be sure that
// the 'cmd' structure contains a valid and up-to-date information.
// You are not allowed to modify the 'cmd' structure.
// Upon entering this function, the 'uFlag' variable contains the flags of
// cmd.ea. If you change the characteristics of the current instruction, you
// are required to refresh 'uFlag'.
// Usually the kernel calls emu() with consecutive addresses in cmd.ea but
// you can't rely on this - for example, if the user asks to analyse an
// instruction at arbirary address, his request will be handled immediately,
// thus breaking the normal sequence of emulation.
// If you need to analyse the surroundings of the current instruction, you
// are allowed to save the contents of the 'cmd' structure and call ana().
// For example, this is a very common pattern:
//  {
//    insn_t saved = cmd;
//    if ( decode_prev_insn(cmd.ea) != BADADDR )
//    {
//      ....
//    }
//    cmd = saved;
//  }
//
// This sample emu() function is a very simple emulation engine.

int emu(void)
{
	int Feature = Instructions[cmd.itype].feature;
	flow = ((Feature & CF_STOP) == 0);

	if (Feature & CF_USE1) handle_operand(cmd.Op1, 1);
	if (Feature & CF_USE2) handle_operand(cmd.Op2, 1);
	if (Feature & CF_USE3) handle_operand(cmd.Op3, 1);
	if (Feature & CF_JUMP) QueueMark(Q_jumps,cmd.ea);
	if (Feature & CF_CHG1) handle_operand(cmd.Op1, 0);
	if (Feature & CF_CHG2) handle_operand(cmd.Op2, 0);
	if (Feature & CF_CHG3) handle_operand(cmd.Op3, 0);

	// if the execution flow is not stopped here, then create
	// a xref to the next instruction.
	// Thus we plan to analyze the next instruction.

	if (flow) ua_add_cref( 0, cmd.ea+cmd.size, fl_F );
	return 1;    // actually the return value is unimportant, but let's it be so
}

//----------------------------------------------------------------------
int is_jump_func(const func_t * /*pfn*/, ea_t *jump_target)
{
	*jump_target = BADADDR;
	return 1; // means "no"
}

//----------------------------------------------------------------------
int may_be_func(void)           // can a function start here?
// arg: none, the instruction is in 'cmd'
// returns: probability 0..100
// 'cmd' structure is filled upon the entrace
// the idp module is allowed to modify 'cmd'
{
	if ( cmd.itype == MSP430_PUSH && cmd.Op1.type == o_reg ) return 50;
	return 0;
}

//----------------------------------------------------------------------
int is_sane_insn(int /*nocrefs*/)
{
	if ( cmd.itype == MSP430_EMU_NOP ) {
		for ( int i=0; i < 8; i++ ) if ( get_word(cmd.ea-i*2) != 0 ) return 1;
		return 0; // too many nops in a row
	}
	return 1;
}

//----------------------------------------------------------------------
int is_align_insn(ea_t ea)
{
	if ( !ua_ana0(ea) ) return 0;
	switch ( cmd.itype )
	{
	case MSP430_EMU_NOP:
		break;
	default:
		return 0;
	}
	return cmd.size;
}
