#include "msp430.hpp"
#include <fpro.h>
#include <diskio.hpp>

#ifdef KR_DEBUG_MESSAGES
//----------------------------------------------------------------------
static void vlog(const char *format, va_list va)
{
	static FILE *fp = NULL;
	if ( fp == NULL )
		fp = fopenWT("debug_log");
	qvfprintf(fp, format, va);
	qflush(fp);
}

//----------------------------------------------------------------------
void log(const char *format, ...)
{
	va_list va;
	va_start(va, format);
	vlog(format, va);
	va_end(va);
}

#else // KR_DEBUG_MESSAGES

void log(const char *format, ...) {}

#endif // KR_DEBUG_MESSAGES

//----------------------------------------------------------------------
inline void outreg( int r )
{
	out_register( ph.regNames[r] );
}

//----------------------------------------------------------------------
static void outmem( op_t &x, ea_t ea )
{
	const char *ptr = get_name_expr(cmd.ea+x.offb, x.n, ea, BADADDR);
	if ( ptr == NULL )
	{
		const ioport_t *p = find_sym(x.addr);
		if ( p == NULL )
		{
			out_tagon(COLOR_ERROR);
			OutLong(x.addr, 16);
			out_tagoff(COLOR_ERROR);
			QueueMark(Q_noName,cmd.ea);
		}
		else
		{
			out_line(p->name, COLOR_IMPNAME);
		}
	}
	else
	{
		bool complex = strchr(ptr, '+') || strchr(ptr, '-');
		if ( complex ) out_symbol(ash.lbrace);
		OutLine(ptr);
		if ( complex ) out_symbol(ash.rbrace);
	}
}

//----------------------------------------------------------------------
bool outop( op_t &op )
{ 
	switch (op.type)
	{

	case o_void:
		return 0;

	case o_reg:
		outreg(op.reg);
		break;

	case o_imm:
		out_symbol('#');
		OutValue(op, OOFS_IFSIGN | OOFW_IMM);
		break;

	case o_displ:
		OutValue(op, OOF_ADDR | OOFS_IFSIGN | OOFW_16);
		out_symbol('(');
		outreg(op.reg);
		out_symbol(')');
		break;

	case o_phrase:
		switch(op.phrase)
		{
	case phrase_ind_reg:
		out_symbol('@');
		outreg(op.indreg);
		break;
	case phrase_ind_reg_inc:
		out_symbol('@');
		outreg(op.indreg);
		out_symbol('+');
		break;
		}
		break;

	case o_mem:
		out_symbol('&');
	case o_near:
		outmem(op, calc_mem(op.addr));
		break;

	default:
		interr("out");
		break;

	}

	return true;
}

//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure
void out( void )
{
	char buf[MAXSTR];
	init_output_buffer(buf, sizeof(buf)); // setup the output pointer

	const char *szPostFix = (cmd.auxpref & CMDAUXPREF_BYTE_FLAG) ? ".b" : NULL;
	OutMnem(8, szPostFix);					// output instruction mnemonics

	if( cmd.Op1.type != o_void) {
		out_one_operand(0);                 // output the src operand
		out_symbol(',');
		OutChar(' ');
	}

	if( cmd.Op2.type != o_void) {
		out_one_operand(1);					// output the dst operand
	}

	if( cmd.Op3.type != o_void) {
		out_symbol(',');
		OutChar(' ');
		out_one_operand(2);                 // output the third operand
	}

	// output a character representation of the immediate values
	// embedded in the instruction as comments
	if (isVoid(cmd.ea,uFlag,0)) OutImmChar(cmd.Op1);
	if (isVoid(cmd.ea,uFlag,1)) OutImmChar(cmd.Op2);	
	if (isVoid(cmd.ea,uFlag,2)) OutImmChar(cmd.Op3);

	term_output_buffer();                 // terminate the output string
	gl_comm = 1;                          // ask to attach a possible user-defined comment to it
	MakeLine(buf);                        // pass the generated line to the
}

//--------------------------------------------------------------------------
// generate start of the disassembly
void header(void)
{
	gen_cmt_line( "Processor:        %s", inf.procName );
	gen_cmt_line( "Device:           %s", device );
	gen_cmt_line( "Target assembler: %s", ash.name );
	if( ash.header != NULL ) {
		for (const char **ptr=ash.header; *ptr != NULL; ptr++) MakeLine(*ptr, 0);
	}
}

//--------------------------------------------------------------------------
// generate start of a segment
void segstart( ea_t ea )
{
	segment_t *Sarea = getseg(ea);
	char *segname = get_segm_name(Sarea);

	if (segname == NULL)
		return;

	if (*segname == '_')
		segname++;

	printf_line(inf.indent, COLSTR("RSEG %s", SCOLOR_ASMDIR), segname);

	unsigned long orgbase = ea - get_segm_para(Sarea);

	if (orgbase != 0) {
		printf_line(-1, COLSTR("%s %s", SCOLOR_ASMDIR),
			ash.origin, btoa(orgbase));
	}
}

//--------------------------------------------------------------------------
// generate end of the disassembly
void footer( void )
{
	char name[MAXSTR];
	get_colored_name(BADADDR, inf.beginEA, name, sizeof(name));
	const char *end = ash.end;
	if ( end == NULL )
		printf_line(-1,COLSTR("%s end %s",SCOLOR_AUTOCMT), ash.cmnt, name);
	else
		printf_line(-1,COLSTR("%s",SCOLOR_ASMDIR)
		" "
		COLSTR("%s %s",SCOLOR_AUTOCMT), ash.end, ash.cmnt, name);
}

//--------------------------------------------------------------------------
// generate a data representation
// usually all the job is handled by the kernel's standard procedure,
// intel_data()
void MSP430_data( ea_t ea )
{
	gl_name = 1;

	// the kernel's standard routine which outputs the data knows nothing
	// about "equ" directives. So we do the following:
	//    - try to output an "equ" directive
	//    - if we succeed, then ok
	//    - otherwise let the standard data output routine, intel_data()
	//        do all the job

	// if ( !out_equ(ea) )
	intel_data(ea);
}	
