#include "msp430.hpp"
#include <entry.hpp>

static const char cfgname[] = "msp430.cfg";

//--------------------------------------------------------------------------
static char *RegNames[] =
{
	"PC","SP","SR","CG2","R4","R5","R6","R7","R8","R9","R10","R11","R12","R13","R14","R15",
	"cs","ds"
};

//--------------------------------------------------------------------------
static uchar retcode_1[] = { 0x30, 0x41 };	// ret
static uchar retcode_2[] = { 0x00, 0x13 };	// reti

static bytes_t retcodes[] = {
	{ sizeof(retcode_1), retcode_1 },
	{ sizeof(retcode_2), retcode_2 },
	{ 0, NULL }                            // NULL terminated array
};

//-----------------------------------------------------------------------
static asm_t std_asm = {
	AS_COLON|ASH_HEXF0|ASB_BINF0|AS_N2CHR|ASO_OCTF0|AS_RELSUP|AS_ASCIIC,
	NULL,
	"MSP430 IAR Assembler",    // name
	0,					// help
	NULL,				// no headers
	NULL,				// no bad instructions
	"ORG",				// ORG directive
	"END",				// END directive
	";",				// comment string
	'"',				// string delimiter
	'\'',				// char delimiter
	"\\\"'",			// special symbols in char and string constants
	"DB",				// ascii string directive
	"DB",				// byte directive
	"DW",				// word directive
	"DC32",				// dword  (4 bytes)
	NULL,				// qword  (8 bytes)
	NULL,				// oword  (16 bytes)
	"DF",				// float  (4 bytes)
	".double",			// double (8 bytes)
	NULL,				// tbyte  (10/12 bytes)
	NULL,				// packed decimal real
	NULL,				// arrays (#h,#d,#v,#s(...)
	"DS %s",			// uninitialized arrays
	"EQU",				// equ
	NULL,				// seg prefix
	NULL,
	NULL,
	NULL,
	NULL,
	"$",				// current ip
	NULL,				// func_header
	NULL,				// func_footer
	"PUBLIC",			// public
	NULL,				// weak
	"EXTERN",			// extrn
	NULL,				// comm
	NULL,				// get_type_name
	"ALIGN",			// align
	'(', ')',			// lbrace, rbrace
	"MOD",				// mod
	"AND",				// and
	"OR",				// or
	"XOR",				// xor
	"NOT",				// not
	"SHL",				// shl
	"SHR",				// shr
	"SIZEOF",			// sizeof
	NULL,				// flag2
	NULL,				// cmnt2
	"LOW",				// low8
	"HIGH",				// high8
	"LWRD",				// low16
	"HWRD"				// high16
};

static asm_t *asms[] = { &std_asm, NULL };

//--------------------------------------------------------------------------
ea_t memstart;

char device[MAXSTR] = "";
static size_t numports;
static ioport_t *ports = NULL;

#include "../iocommon.cpp"

//----------------------------------------------------------------------
const ioport_t *find_sym(int address)
{
	const ioport_t *port = find_ioport(ports, numports, address);
	return port;
}

//----------------------------------------------------------------------
static void create_words(void)
{
	for ( int i=0; i < numports; i++ ) {
		ea_t ea = ports[i].address;
		if ( isTail(getFlags(ea)) ) do_unknown(ea, false);
		doWord(ea, 2);
	}
}

//----------------------------------------------------------------------
bool msp430_select_ioport( void )
{
	return ( choose_ioport_device(cfgname, device) );                                   // IDA 4.70
	// return ( choose_ioport_device(cfgname, device, sizeof(device), parse_area_line0) ); // IDA 4.80
}

//--------------------------------------------------------------------------
const char *set_idp_options(const char *keyword,int /*value_type*/,const void * /*value*/)
{
	if ( keyword != NULL ) return IDPOPT_BADKEY;
	if ( msp430_select_ioport() ) set_device_name(device, IORESP_PORT|IORESP_INT);
	return IDPOPT_OK;
}

//--------------------------------------------------------------------------
netnode helper;

static int notify(processor_t::idp_notify msgid, ...)
{
	va_list va;
	va_start(va, msgid);

	// A well behaving processor module should call invoke_callbacks()
	// in his notify() function. If this function returns 0, then
	// the processor module should process the notification itself
	// Otherwise the code should be returned to the caller:

	int code = invoke_callbacks(HT_IDP, msgid, va);
	if ( code ) return code;

	switch(msgid)
	{
	case ph.init:
		helper.create("$ msp430");
		inf.mf = 0;       // Set a little endian mode of the IDA kernel
		break;

	case ph.newfile:   // new file loaded
		if ( msp430_select_ioport() ) set_device_name(device, IORESP_ALL);
		create_words();
		break;

	case ph.oldfile:   // old file loaded
		set_device_name(helper.supval(-1), IORESP_NONE);
		break;

	case ph.is_jump_func:
		{
			const func_t *pfn = va_arg(va, const func_t *);
			ea_t *jump_target = va_arg(va, ea_t *);
			return is_jump_func(pfn, jump_target);
		}

	case ph.is_sane_insn:
		return is_sane_insn(va_arg(va, int));

	case ph.may_be_func:        // can a function start here?
		// arg: none, the instruction is in 'cmd'
		// returns: probability 0..100
		// 'cmd' structure is filled upon the entrace
		// the idp module is allowed to modify 'cmd'
		return may_be_func();

	}
	va_end(va);
	return 1;
}

static char *shnames[] = { "msp430", NULL };
static char *lnames[]  = { "Texas Instruments MSP430", NULL };

//-----------------------------------------------------------------------
//      Processor Definition
//-----------------------------------------------------------------------
processor_t LPH =
{
	IDP_INTERFACE_VERSION,// version
	0x8001,	            // id
	PR_RNAMESOK         // can use register names for byte names
	|PR_BINMEM,         // The module creates RAM/ROM segments for binary files
	// (the kernel shouldn't ask the user about their sizes and addresses)
	8,                  // 8 bits in a byte for code segments
	8,                  // 8 bits in a byte for other segments

	shnames,            // array of short processor names
	// the short names are used to specify the processor
	// with the -p command line switch)
	lnames,             // array of long processor names
	// the long names are used to build the processor type
	// selection menu

	asms,               // array of target assemblers

	notify,             // the kernel event notification callback

	header,             // generate the disassembly header
	footer,             // generate the disassembly footer

	segstart,           // generate a segment declaration (start of segment)
	std_gen_segm_footer,// generate a segment footer (end of segment)

	NULL,               // generate 'assume' directives

	ana,                // analyze an instruction and fill the 'cmd' structure
	emu,                // emulate an instruction

	out,                // generate a text representation of an instruction
	outop,              // generate a text representation of an operand
	MSP430_data,        // generate a text representation of a data item
	NULL,               // compare operands
	NULL,               // can an operand have a type?

	qnumber(RegNames),  // Number of registers
	RegNames,           // Register names
	NULL,               // get abstract register

	0,                  // Number of register files
	NULL,               // Register file names
	NULL,               // Register descriptions
	NULL,               // Pointer to CPU registers

	rVcs,               // first
	rVds,               // last
	2,                  // size of a segment register
	rVcs, rVds,

	NULL,               // No known code start sequences
	retcodes,

	0,MSP430_last,
	Instructions,
	NULL,
	NULL,
	NULL,
	NULL,
	{NULL,NULL,NULL,NULL},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	gen_spcdef,
	MSP430_EMU_RET,
	set_idp_options,
	is_align_insn,
	NULL
};

